/*
 * yasapi_cache.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

static char *_wcs2strdup(wchar_t *wcs, const char *pfx, const char *sfx,
    int bMarkup);

int CacheCreate(Cache *pCache, Player *pPlayer)
{
  HRESULT hr;
  UINT nDevice;

  /////////////////////////////////////////////////////////////////////////////
  pCache->pPlayer=pPlayer;

  /////////////////////////////////////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,      // _In_  REFCLSID  rclsid,
    NULL,                           // _In_  LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                     // _In_  DWORD     dwClsContext,
    &IID_IMMDeviceEnumerator,       // _In_  REFIID    riid,
    (void**)&pCache->pEnumerator   // _Out_ LPVOID    *ppv
  );

  if (FAILED(hr)) switch (hr) {
  case REGDB_E_CLASSNOTREG:
    DMESSAGE("CoCreateInstance() failed with REGDB_E_CLASSNOTREG");
    goto eenumerator;
  case CLASS_E_NOAGGREGATION:
    DMESSAGE("CoCreateInstance() failed with CLASS_E_NOAGGREGATION");
    goto eenumerator;
  case E_NOINTERFACE:
    DMESSAGE("CoCreateInstance() failed with E_NOINTERFACE");
    goto eenumerator;
  case E_POINTER:
    DMESSAGE("CoCreateInstance() failed with E_POINTER");
    goto eenumerator;
  case 0x800401f0:
    DMESSAGE("CoCreateInstance() failed with COM not initialized");
    goto eenumerator;
  default:
    DVMESSAGE("CoCreateInstance() failed with 0x%p/%lu",hr,hr);
    goto eenumerator;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pCache->pEnumerator->lpVtbl->EnumAudioEndpoints(pCache->pEnumerator,
    eRender,                        // [in]  EDataFlow           dataFlow,
    DEVICE_STATE_ACTIVE,            // [in]  DWORD               dwStateMask,
    &pCache->pCollection            // [out] IMMDeviceCollection **ppDevices
  );

  if (FAILED(hr)) switch (hr) {
  case E_POINTER:
    DMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with"
        " E_POINTER");
    goto ecollection;
  case E_INVALIDARG:
    DMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with"
        " E_INVALIDARG");
    goto ecollection;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with"
        " E_OUTOFMEMORY");
    goto ecollection;
  default:
    DVMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with %p",hr);
    goto ecollection;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pCache->pCollection->lpVtbl->GetCount(pCache->pCollection,
    &pCache->cCacheDevices         // [out] UINT *pcDevices
  );

  if (FAILED(hr)) switch (hr) {
  case E_POINTER:
    DMESSAGE("IMMDeviceCollection::GetCount failed with E_POINTER");
    goto ecount;
  case E_INVALIDARG:
    DMESSAGE("IMMDeviceCollection::GetCount failed with E_INVALIDARG");
    goto ecount;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDeviceCollection::GetCount failed with E_OUTOFMEMORY");
    goto ecount;
  default:
    DVMESSAGE("IMMDeviceCollection::GetCount() failed with %p",hr);
    goto ecount;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (OptionsCreate(&pCache->options,pPlayer,pCache->cCacheDevices)<0) {
    DMESSAGE("Creating options.");
    goto eoptions;
  }

  /////////////////////////////////////////////////////////////////////////////
  ++pCache->cCacheDevices;
  pCache->pCacheDevices
      =malloc(pCache->cCacheDevices*sizeof *pCache->pCacheDevices);

  if (!pCache->pCacheDevices) {
    DMESSAGE("allocating CacheDevice array");
    goto emalloc;
  }

  for (nDevice=0;nDevice<pCache->cCacheDevices;++nDevice) {
    if (CacheDeviceCreate(pCache->pCacheDevices+nDevice,pCache,nDevice-1)<0) {
      DMESSAGE("creating CacheDevice");
      goto edevice;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (CacheSelectDevice(pCache,NULL)<0) {
    DMESSAGE("selecting device");
    goto eselect;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//error:
eselect:
edevice:
  while (0<nDevice) {
    --nDevice;
    CacheDeviceDestroy(pCache->pCacheDevices+nDevice,&pCache->options);
  }

  pCache->cCacheDevices=0u;
  free(pCache->pCacheDevices);
emalloc:
  OptionsDestroy(&pCache->options);
eoptions:
ecount:
  pCache->pCollection->lpVtbl->Release(pCache->pCollection);
ecollection:
  pCache->pEnumerator->lpVtbl->Release(pCache->pEnumerator);
eenumerator:
  return -1;
}

void CacheDestroy(Cache *pCache)
{
  UINT nDevice=pCache->cCacheDevices;

  while (0u<nDevice) {
    --nDevice;
    CacheDeviceDestroy(pCache->pCacheDevices+nDevice,&pCache->options);
  }

  free(pCache->pCacheDevices);
  OptionsDestroy(&pCache->options);
  pCache->pCollection->lpVtbl->Release(pCache->pCollection);
  pCache->pEnumerator->lpVtbl->Release(pCache->pEnumerator);
}

int CacheSelectDevice(Cache *pCache, CacheDevice *pCacheDevice)
{
  wchar_t *lhs=pCache->options.common.current.id;
  const wchar_t *rhs;
  UINT nDevice;

  if (pCacheDevice) {
    //pCache->selected.active.pCacheDevice=pCacheDevice;
    pCache->selected.config.pCacheDevice=pCacheDevice;
    rhs=pCacheDevice->lpVtbl->getid(pCacheDevice);
    wcscpy(lhs,rhs);
  }
  else {
    if (0u==pCache->cCacheDevices) {
      DMESSAGE("no devices available");
      pCache->selected.active.pCacheDevice=NULL;
      pCache->selected.config.pCacheDevice=NULL;
      goto error;;
    }

    for (nDevice=0u;nDevice<pCache->cCacheDevices;++nDevice) {
      pCacheDevice=pCache->pCacheDevices+nDevice;
      rhs=pCacheDevice->lpVtbl->getid(pCacheDevice);

      if (!wcscmp(lhs,rhs)) {
        pCache->selected.active.pCacheDevice=pCacheDevice;
        pCache->selected.config.pCacheDevice=pCacheDevice;
        goto success;
      }
    }

    pCacheDevice=pCache->pCacheDevices;
    pCache->selected.active.pCacheDevice=pCacheDevice;
    pCache->selected.config.pCacheDevice=pCacheDevice;
    rhs=pCacheDevice->lpVtbl->getid(pCacheDevice);
    wcscpy(lhs,rhs);
  }
success:
  DVWRITELNW(L"selected=\"%s\"",lhs);
  return 0;
error:
  return -1;
}

void CacheRestore(Cache *pCache)
{
  UINT nDevice;
  CacheDevice *pDevice;

  pCache->options.common.config=pCache->options.common.current;

  for (nDevice=0u;nDevice<pCache->cCacheDevices;++nDevice) {
    pDevice=pCache->pCacheDevices+nDevice;
    pDevice->pMember->config=pDevice->pMember->current;
  }

  pCache->selected.config.pCacheDevice=pCache->selected.active.pCacheDevice;
}

void CacheSync(Cache *pCache)
{
  Player *pPlayer=pCache->pPlayer;
  UINT nDevice;
  CacheDevice *pDevice;

  if (FsmState(&pPlayer->present)<ePresentClientStateRunning)
    goto copy;

  YASAPI_FSM_BACKTRACK(pPlayer->present.pServer,ePresentServerStateTimerStart);
copy:
  pCache->options.common.current=pCache->options.common.config;

  for (nDevice=0u;nDevice<pCache->cCacheDevices;++nDevice) {
    pDevice=pCache->pCacheDevices+nDevice;
    pDevice->pMember->current=pDevice->pMember->config;
  }

  pCache->selected.active.pCacheDevice=pCache->selected.config.pCacheDevice;
}

void CacheDefault(Cache *pCache)
{
  UINT nDevice;
  CacheDevice *pDevice;

  pCache->options.common.config=gcOptionsCommon;

  for (nDevice=0u;nDevice<pCache->cCacheDevices;++nDevice) {
    pDevice=pCache->pCacheDevices+nDevice;
    pDevice->pMember->config=gcOptionsDevice;
  }

  pCache->selected.config.pCacheDevice=pCache->pCacheDevices;
}

///////////////////////////////////////////////////////////////////////////////
static const wchar_t *WasaspiDeviceGetId(CacheDevice *pDevice)
{
  return pDevice->id;
}

static ICacheDeviceVtbl gcICacheDeviceVtbl={
  .getid=WasaspiDeviceGetId,
};

///////////////////////////////////////////////////////////////////////////////
static const wchar_t *WasaspiDeviceDefaultGetId(CacheDevice *pDevice)
{
  return L"";
}

static ICacheDeviceVtbl gcICacheDeviceVtblDefault={
  .getid=WasaspiDeviceDefaultGetId,
};

///////////////////////////////////////////////////////////////////////////////
int CacheDeviceCreate(CacheDevice *pDevice, Cache *pCache, int nDevice)
{
  HRESULT hr;

  if (nDevice<0) {
    ///////////////////////////////////////////////////////////////////////////
    pDevice->lpVtbl=&gcICacheDeviceVtblDefault;

    hr=pCache->pEnumerator->lpVtbl->GetDefaultAudioEndpoint(
      pCache->pEnumerator,
      eRender,                      // [in]  EDataFlow dataFlow,
      eConsole,                     // [in]  ERole     role,
      &pDevice->pDevice             // [out] IMMDevice **ppDevice
    );

    if (FAILED(hr))
#if defined (_MSC_VER) // [
      if (E_NOTFOUND==hr) {
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_NOTFOUND");
        goto edevice;
      }
      else
#endif
      switch (hr) {
      case E_POINTER:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_POINTER");
        goto edevice;
      case E_INVALIDARG:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_INVALIDARG");
        goto edevice;
#if ! defined (_MSC_VER) // [
      case E_NOTFOUND:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_NOTFOUND");
        goto edevice;
#endif // ]
      case E_OUTOFMEMORY:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_OUTOFMEMORY");
        goto edevice;
      default:
        DVMESSAGE("IIMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with %p",hr);
        goto edevice;
      }
  }
  else {
    ///////////////////////////////////////////////////////////////////////////
    pDevice->lpVtbl=&gcICacheDeviceVtbl;

    ///////////////////////////////////////////////////////////////////////////
    hr=pCache->pCollection->lpVtbl->Item(pCache->pCollection,
      nDevice,                      // [in]  UINT      nDevice,
      &pDevice->pDevice             // [out] IMMDevice **ppDevice
    );

    if (FAILED(hr)) switch (hr) {
    case E_POINTER:
      DMESSAGE("IMMDeviceCollection::Item() failed with E_POINTER");
      goto edevice;
    case E_INVALIDARG:
      DMESSAGE("IMMDeviceCollection::Item() failed with E_INVALIDARG");
      goto edevice;
    default:
      DVMESSAGE("IMMDeviceCollection::Item() failed with 0x%p",hr);
      goto edevice;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pDevice->pDevice->lpVtbl->GetId(pDevice->pDevice,
    &pDevice->id                    // [out] LPWSTR *ppstrId
  );

  if (FAILED(hr)) switch (hr) {
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDevice::GetId() failed with E_OUTOFMEMORY");
    goto eid;
  case E_POINTER:
    DMESSAGE("IMMDevice::GetId() failed with E_POINTER");
    goto eid;
  default:
    DVMESSAGE("IMMDevice::GetId() failed with 0x%p",hr);
    goto eid;
  }

  /////////////////////////////////////////////////////////////////////////////
  pDevice->pMember=OptionsSearch(&pCache->options,pDevice->id);

  if (!pDevice->pMember) {
    DMESSAGE("options not found");
    goto emember;
  }

  OptionsMemberDeviceRef(pDevice->pMember);

  /////////////////////////////////////////////////////////////////////////////
  hr=pDevice->pDevice->lpVtbl->OpenPropertyStore(pDevice->pDevice,
    STGM_READ,                      // [in]  DWORD          stgmAccess,
    &pDevice->pStore                // [out] IPropertyStore **ppProperties
  );

  if (FAILED(hr)) switch (hr) {
  case E_INVALIDARG:
    DMESSAGE("IMMDevice::OpenPropertyStore() failed with E_INVALIDARG");
    goto estore;
  case E_POINTER:
    DMESSAGE("IMMDevice::OpenPropertyStore() failed with E_POINTER");
    goto estore;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDevice::OpenPropertyStore() failed with E_OUTOFMEMORY");
    goto estore;
  default:
    DVMESSAGE("IMMDevice::OpenPropertyStore() failed with %p",hr);
    goto estore;
  }

  /////////////////////////////////////////////////////////////////////////////
  PropVariantInit(&pDevice->vFriendlyName);

  hr=pDevice->pStore->lpVtbl->GetValue(pDevice->pStore,
    &PKEY_Device_FriendlyName,      // [in]  REFPROPERTYKEY key,
    &pDevice->vFriendlyName         // [out] PROPVARIANT    *pv
  );

  if (FAILED(hr)) {
    DVMESSAGE("IPropertyStore::GetValue() failed with %p",hr);
    goto efriendly;
  }

  /////////////////////////////////////////////////////////////////////////////
  pDevice->pFriendlyNameOriginal=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
      NULL,NULL,0);

  if (!pDevice->pFriendlyNameOriginal) {
    DMESSAGE("converting FriendlyName to UTF-8");
    goto eoriginal;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (nDevice<0) {
    pDevice->pFriendlyName=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        "Default Device [","]",0);
    pDevice->pFriendlyNameMarkup=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        "Default Device [","]",1);
  }
  else {
    pDevice->pFriendlyName=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        NULL,NULL,0);
    pDevice->pFriendlyNameMarkup=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        NULL,NULL,1);
  }

  if (!pDevice->pFriendlyName) {
    DMESSAGE("converting FriendlyName to UTF-8");
    goto eclone;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (1u==pDevice->pMember->uRef) {
    if (FormatMatrixCreate(&pDevice->pMember->matrix,pDevice->pDevice)<0) {
      DMESSAGE("creating matrix");
      goto ematrix;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//error:
  if (1u==pDevice->pMember->uRef)
    FormatMatrixDestroy(&pDevice->pMember->matrix);
ematrix:
  free(pDevice->pFriendlyName);
eclone:
  free(pDevice->pFriendlyNameOriginal);
eoriginal:
  PropVariantClear(&pDevice->vFriendlyName);
efriendly:
  pDevice->pStore->lpVtbl->Release(pDevice->pStore);
estore:
  OptionsMemberDeviceUnref(pDevice->pMember,&pCache->options);
emember:
  CoTaskMemFree(pDevice->id);
eid:
  pDevice->pDevice->lpVtbl->Release(pDevice->pDevice);
edevice:
  return -1;
}

void CacheDeviceDestroy(CacheDevice *pDevice, Options *pOptions)
{
  if (0u==--pDevice->pMember->uRef)
    FormatMatrixDestroy(&pDevice->pMember->matrix);

  free(pDevice->pFriendlyName);
  free(pDevice->pFriendlyNameOriginal);
  PropVariantClear(&pDevice->vFriendlyName);
  pDevice->pStore->lpVtbl->Release(pDevice->pStore);
  OptionsMemberDeviceUnref(pDevice->pMember,pOptions);
  CoTaskMemFree(pDevice->id);
  pDevice->pDevice->lpVtbl->Release(pDevice->pDevice);
}

///////////////////////////////////////////////////////////////////////////////
int FormatMatrixCreate(FormatMatrix *pMatrix, IMMDevice *pDevice)
{
  WAVEFORMATEX *pClosestMatch=NULL;
  HRESULT hr;
  IAudioClient *pClient;
  int row,col;
  WAVEFORMATEXTENSIBLE wfx;

  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,  // [in]  REFIID      iid,
    CLSCTX_ALL,         // [in]  DWORD       dwClsCtx,
    NULL,               // [in]  PROPVARIANT *pActivationParams,
    (void **)&pClient   // [out] void        **ppInterface
  );
  
  if (FAILED(hr)) switch (hr) {
  case E_NOINTERFACE:
    DMESSAGE("IMMDevice::Activate() failed with E_NOINTERFACE");
    goto eactivate;
  case E_POINTER:
    DMESSAGE("IMMDevice::Activate() failed with E_POINTER");
    goto eactivate;
  case E_INVALIDARG:
    DMESSAGE("IMMDevice::Activate() failed with E_INVALIDARG");
    goto eactivate;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDevice::Activate() failed with E_OUTOFMEMORY");
    goto eactivate;
  case AUDCLNT_E_DEVICE_INVALIDATED:
    DMESSAGE("IMMDevice::Activate() failed with AUDCLNT_E_DEVICE_INVALIDATED");
    goto eactivate;
  default:
    DVMESSAGE("IMMDevice::Activate() failed with %p",hr);
    goto eactivate;
  }

  wfx.Format.wFormatTag=WAVE_FORMAT_EXTENSIBLE;
  wfx.Format.nChannels=2;
  wfx.Format.cbSize=(sizeof wfx)-sizeof (wfx.Format);
  ////
  wfx.dwChannelMask=SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT;
  wfx.SubFormat=KSDATAFORMAT_SUBTYPE_PCM;

  for (row=0;row<eFrequenciesNum;++row) {
    wfx.Format.nSamplesPerSec=gaFrequencies[row].freq;;

    for (col=0;col<eBitDepthsNum;++col) {
      wfx.Format.wBitsPerSample=gaBitDepths[col].bits;
      wfx.Format.nBlockAlign=wfx.Format.nChannels*wfx.Format.wBitsPerSample;
      wfx.Format.nAvgBytesPerSec
          =wfx.Format.nSamplesPerSec*wfx.Format.nBlockAlign;
      ////
      wfx.Samples.wValidBitsPerSample=wfx.Format.wBitsPerSample;

      hr=pClient->lpVtbl->IsFormatSupported(pClient,
                        // [in]        AUDCLNT_SHAREMODE ShareMode,
        AUDCLNT_SHAREMODE_EXCLUSIVE,
        &wfx.Format,    // [in]  const WAVEFORMATEX      *pFormat,
        &pClosestMatch  // [out]       WAVEFORMATEX      **ppClosestMatch
      );

      if (pClosestMatch) {
        CoTaskMemFree(pClosestMatch);
        pClosestMatch=NULL;
      }

      pMatrix->bSupported[row][col]=0;

      switch (hr) {
      case S_OK:
        pMatrix->bSupported[row][col]=1;
        break;
      case S_FALSE:
      case AUDCLNT_E_UNSUPPORTED_FORMAT:
        break;
      case E_POINTER:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " E_POINTER");
        break;
      case E_INVALIDARG:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " E_INVALIDARG");
        break;
      case AUDCLNT_E_DEVICE_INVALIDATED:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " AUDCLNT_E_DEVICE_INVALIDATED");
        break;
      case AUDCLNT_E_SERVICE_NOT_RUNNING:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " AUDCLNT_E_SERVICE_NOT_RUNNING");
        break;
      default:
        DVMESSAGE("IAudioClient::IsFormatSupported() failed with %p",hr);
        break;
      }
    }
  }

  pClient->lpVtbl->Release(pClient);

  return 0;
eactivate:
  return -1;
}

void FormatMatrixDestroy(FormatMatrix *pMatrix)
{
}

///////////////////////////////////////////////////////////////////////////////
static char *_wcs2strdup(wchar_t *wcs, const char *pfx, const char *sfx,
    int bMarkup)
{
  const char BL[]="<b>";
  const char BR[]="</b>";

  enum {
    eLenBL = (sizeof BL) - 1,
    eLenBR = (sizeof BR) - 1,
    eLenB=eLenBL+eLenBR,
  };

  int nLen1=pfx?strlen(pfx):0;
  int nSize2;
  int nLen3=sfx?strlen(sfx):0;
  int nSize;
  char *str,*p;

  if (!wcs)
    goto ewcs;

  nSize2=WideCharToMultiByte(
    CP_UTF8,        // _In_      UINT    CodePage,
    0ul,            // _In_      DWORD   dwFlags,
    wcs,            // _In_      LPCWSTR lpWideCharStr,
    -1,             // _In_      int     cchWideChar,
    NULL,           // _Out_opt_ LPSTR   lpMultiByteStr,
    0,              // _In_      int     cbMultiByte,
    NULL,           // _In_opt_  LPCSTR  lpDefaultChar,
    NULL            // _Out_opt_ LPBOOL  lpUsedDefaultChar
  );

  nSize=nLen1+nSize2+nLen3;
  if (bMarkup)
    nSize+=eLenB;

  if (!(str=malloc(nSize)))
    goto emalloc;

  p=str;

  if (bMarkup) {
    strcpy(p,BL);
    p+=eLenBL;
  }

  if (0<nLen1) {
    strcpy(p,pfx);
    p+=nLen1;
  }

  WideCharToMultiByte(
    CP_UTF8,        // _In_      UINT    CodePage,
    0ul,            // _In_      DWORD   dwFlags,
    wcs,            // _In_      LPCWSTR lpWideCharStr,
    -1,             // _In_      int     cchWideChar,
    p,              // _Out_opt_ LPSTR   lpMultiByteStr,
    nSize2,         // _In_      int     cbMultiByte,
    NULL,           // _In_opt_  LPCSTR  lpDefaultChar,
    NULL            // _Out_opt_ LPBOOL  lpUsedDefaultChar
  );

  p+=nSize2-1;

  if (0<nLen3) {
    strcpy(p,sfx);
    p+=nLen3;
  }

  if (bMarkup) {
    strcpy(p,BR);
    p+=eLenBR;
  }

  return str;
emalloc:
ewcs:
  return NULL;
}
