/*
 * yasapi_config_label_frames.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>
#include <yasapi.h>

static void LabelEngineClientFramesSet(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  Player *pPlayer=pDialog->pGtk->pPlayer;
  EngineClient *pClient=&pPlayer->engine;
  UINT32 *puFrames;
  double xDuration;
  char buf[128];

  if (FsmState(pClient)<eEngineClientStateAudioClient) {
    gtk_label_set_text (GTK_LABEL (pWidget->widget), "");
    goto exit;
  }

  if (NavigateEngineClient!=pWidget->pCell->nav) {
    gtk_label_set_text (GTK_LABEL (pWidget->widget), "");
    goto exit;
  }

  puFrames=(UINT32 *)pWidget->pCell->nav(pDialog,pWidget->pCell->offs);

  if (!puFrames||UINT32_MAX==*puFrames) {
    gtk_label_set_text (GTK_LABEL (pWidget->widget), "");
    goto exit;
  }

  xDuration=YASAPI_FRAMES2MS(*puFrames,&pClient->wfx.target);

  sprintf(buf,"%u Frames@%lu Hz ≈ %.0f ms ",*puFrames,
      pClient->wfx.target.Format.nSamplesPerSec,xDuration);
  gtk_label_set_text (GTK_LABEL (pWidget->widget), buf);
exit:
  return;
}

static void LabelEngineClientFramesGet(ConfigWidget *pWidget)
{
}

///////////////////////////////////////////////////////////////////////////////
static GtkWidget *ClassLabelEngineClientFramesCreate(ClassCell *pCell,
    Dialog *pDialog, Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
    .set=LabelEngineClientFramesSet,
    .get=LabelEngineClientFramesGet,
  };

  GtkWidget *widget = gtk_label_new(NULL);

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);

  return widget;
}

IClassCellVtbl gcIClassLabelEngineClientFramesVtbl={
  .create=ClassLabelEngineClientFramesCreate,
};
