/*
 * yasapi_consumer_time.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID_TIME(id) \
  YASAPI_PASTE3(EngineServerTime,YASAPI_ID,id)
#define YASAPI_ID_ITIME_VTBL(id) \
  YASAPI_PASTE3(gcIEngineServerTime,id,Vtbl)

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Null

static int YASAPI_ID_TIME(ClockCreate)(EngineServer *pServer)
{
  return 0;
}

static void YASAPI_ID_TIME(ClockDestroy)(EngineServer *pServer)
{
}

static void YASAPI_ID_TIME(Inc)(EngineServer *pServer)
{
}

static int YASAPI_ID_TIME(Get)(EngineServer *pServer)
{
  return 0;
}

static void YASAPI_ID_TIME(Set)(EngineServer *pServer, int ms)
{
}

IEngineServerTimeVtbl YASAPI_ID_ITIME_VTBL(YASAPI_ID)={
  .pId=YASAPI_STRA(YASAPI_ID),
  .ClockCreate=YASAPI_ID_TIME(ClockCreate),
  .ClockDestroy=YASAPI_ID_TIME(ClockDestroy),
  .Inc=YASAPI_ID_TIME(Inc),
  .Get=YASAPI_ID_TIME(Get),
  .Set=YASAPI_ID_TIME(Set),
};

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Clock

static int YASAPI_ID_TIME(ClockCreate)(EngineServer *pServer)
{
  IAudioClient *pAudioClient=pServer->pAudioClient;
  HRESULT hr;

  DASSERT(pAudioClient,exit);

  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioClock,         // [in]  REFIID riid,
    (void **)&pServer->pAudioClock
                              // [out] void   **ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",E_POINTER);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",E_NOINTERFACE);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClient::GetService(IAudioClock)",hr);
    }

    goto exit;
  }

  return 0;
exit:
  return -1;
}

static void YASAPI_ID_TIME(ClockDestroy)(EngineServer *pServer)
{
  IAudioClock **ppAudioClock=&pServer->pAudioClock;

  (*ppAudioClock)->lpVtbl->Release(*ppAudioClock);
  *ppAudioClock=NULL;
}

static double _EngineServerTimeClockGet(EngineServer *pServer)
{
  double ms=0.0;
  Player *pPlayer=pServer->pClient->pPlayer;
  IAudioClock *pAudioClock=pServer->pAudioClock;
  EEngineServer eEngineServer;
  UINT64 u64Frequency;
  UINT64 u64Position;
  HRESULT hr;

  if (FsmState(pServer)<eEngineServerStateBase)
    goto ebase;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  eEngineServer=FsmState(pServer);

  switch (eEngineServer) {
  default:
    if (eEngineServer<eEngineServerStateAudioClock)
      goto exit;

    break;
  }

  hr=pAudioClock->lpVtbl->GetFrequency(pAudioClock,
    &u64Frequency     // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetFrequency",E_POINTER);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetFrequency",hr);
    }

    goto exit;
  }

  hr=pAudioClock->lpVtbl->GetPosition(pAudioClock,
    &u64Position,     // [out] UINT64 *pu64Position
    NULL              // UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetPosition",E_POINTER);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetPosition",hr);
    }

    goto exit;
  }

  ms=1000.0*u64Position/u64Frequency;
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
ebase:
  return ms;
}

static void YASAPI_ID_TIME(Inc)(EngineServer *pServer)
{
  pServer->pClient->param.time.ms=_EngineServerTimeClockGet(pServer);
}

static int YASAPI_ID_TIME(Get)(EngineServer *pServer)
{
  return _EngineServerTimeClockGet(pServer)-pServer->pClient->param.time.ms+0.5;
}

static void YASAPI_ID_TIME(Set)(EngineServer *pServer, int ms)
{
  pServer->pClient->param.time.ms=_EngineServerTimeClockGet(pServer)-ms;
}

IEngineServerTimeVtbl YASAPI_ID_ITIME_VTBL(YASAPI_ID)={
  .pId=YASAPI_STRA(YASAPI_ID),
  .ClockCreate=YASAPI_ID_TIME(ClockCreate),
  .ClockDestroy=YASAPI_ID_TIME(ClockDestroy),
  .Inc=YASAPI_ID_TIME(Inc),
  .Get=YASAPI_ID_TIME(Get),
  .Set=YASAPI_ID_TIME(Set),
};
