/*
 * yasapi_producer_wfx.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

int EngineWfxCreate(EngineWfx *pWfx, const EngineOptions *pOptions,
		int samplerate, int numchannels, int bitspersamp)
{
	int code=-1;
	WAVEFORMATEXTENSIBLE *pSource=&pWfx->source;
	WAVEFORMATEXTENSIBLE *pTarget=&pWfx->target;
  int bSurround=pOptions->common.bSurround;
  WORD nChannelsTarget;
  WORD wBitsPerSampleTarget;

  /////////////////////////////////////////////////////////////////////////////
	memset(pWfx,0,sizeof *pWfx);

  /////////////////////////////////////////////////////////////////////////////
  if (WfxCreate(pSource,samplerate,numchannels,bitspersamp,bSurround)<0) {
    DVMESSAGE("unknown format %d/%d/%d",samplerate,numchannels,
        bitspersamp);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (1==numchannels&&pOptions->common.eMono2Stereo)
    nChannelsTarget=2;
  else
    nChannelsTarget=numchannels;

  switch (pOptions->device.eBitdepth) {
  case eBitdepth16:
    wBitsPerSampleTarget=16u;
    break;
  case eBitdepth24:
    wBitsPerSampleTarget=24u;
    break;
  case eBitdepth32:
    wBitsPerSampleTarget=32u;
    break;
  default:
    wBitsPerSampleTarget=bitspersamp;
    break;
  }

  DVWRITELN("wBitsPerSampleTarget: %u",wBitsPerSampleTarget);

  /////////////////////////////////////////////////////////////////////////////
  if (WfxCreate(pTarget,samplerate,nChannelsTarget,wBitsPerSampleTarget,
      bSurround)<0) {
    DVMESSAGE("unknown format %d/%u/%u",samplerate,nChannelsTarget,
        wBitsPerSampleTarget);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
	code=0;
exit:
	return code;
}
