/*
 * yasapi_player.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <winamp/wa_ipc.h>

///////////////////////////////////////////////////////////////////////////////
// http://forums.winamp.com/archive/index.php/t-316807.html
#define IPC_GETINIFILEW 1334
#define IPC_GETINIDIRECTORYW 1335
#define IPC_GETPLUGINDIRECTORYW 1336
#define IPC_GETVISDIRECTORYW 339
#define IPC_GETSKINDIRECTORYW 340
#define IPC_GETDSPDIRECTORYW 341

///////////////////////////////////////////////////////////////////////////////
int PlayerCreate(Player *pPlayer, Out_Module *pOut)
{
  int code=-1;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(pPlayer,sizeof *pPlayer);


  FsmCreate(pPlayer,gcaIPlayerState ,NULL,NULL);

  pPlayer->stats.uNumTracks=UINT32_MAX;
  pPlayer->stats.uNumGapless=UINT32_MAX;
#if defined (YASAPI_BROKEN_DISPLAY) || ! defined (YASAPI_NBROKEN) // [
  pPlayer->stats.uNumGaplessBroken=UINT32_MAX;
#endif // ]
  YASAPI_DOUBLE_INVALIDATE(pPlayer->stats.xMsDropout);

  EngineStatsReset(&pPlayer->engine.stats);

  pPlayer->pOut=pOut;

  GetModuleFileNameW(
    pOut->hDllInstance,     // _In_opt_ HMODULE hModule,
    pPlayer->path,          // _Out_    LPTSTR  lpFilename,
    (sizeof pPlayer->path)/sizeof (pPlayer->path[0])
                            // _In_     DWORD   nSize
  );

  pPlayer->pBasename=yasapi_basenamew(pPlayer->path);

  pPlayer->ini[0]=0;
  pPlayer->param.nVolume=ePlayerVolumeMax;
  pPlayer->stats.uMsLatencyMax=UINT32_MAX;

  /////////////////////////////////////////////////////////////////////////////
  if (EngineClientCreate(&pPlayer->engine,pPlayer)<0) {
    DMESSAGE("creating engine client");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (GtkClientCreate(&pPlayer->gtk,pPlayer)<0) {
    DMESSAGE("creating gtk client");
    goto exit;
  }

  if (PresentClientCreate(&pPlayer->present,pPlayer)<0) {
    DMESSAGE("creating present client");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (YASAPI_FSM_UP(pPlayer,ePlayerStateCriticalSection)<0) {
    DMESSAGE("something went wrong unexpectedly");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:

  return code;
}

void PlayerDestroy(Player *pPlayer)
{
  PresentClientDestroy(&pPlayer->present);
  GtkClientDestroy(&pPlayer->gtk);
  EngineClientDestroy(&pPlayer->engine);
  YASAPI_FSM_DOWN(pPlayer,0);
}

///////////////////////////////////////////////////////////////////////////////
int PlayerPropertiesPath(Player *pPlayer)
{
  static const wchar_t SUBDIR[]=L"\\Plugins";

  int code=-1;
  Out_Module *pOut=pPlayer->pOut;
  wchar_t *ini=pPlayer->ini;
  wchar_t *inidir,*ep,*pp,*wp;
  DWORD dwSize;

  /////////////////////////////////////////////////////////////////////////////
  ini[0]=0;

  if (!pOut)
    goto eout;

  /////////////////////////////////////////////////////////////////////////////
  // extension pointer:
  dwSize=wcslen(pPlayer->path)+1;

  for (ep=pPlayer->path+dwSize;pPlayer->path<ep&&L'.'!=ep[-1];--ep)
    ;

  // path pointer:
  for (pp=ep;pPlayer->path<pp&&L'\\'!=pp[-1];--pp)
    ;

  // write pointer:
  wp=ini;

  // copy the path.
  inidir=(wchar_t *)SendMessageW(pOut->hMainWindow,WM_WA_IPC,0,
      IPC_GETINIDIRECTORYW);
  wcscpy(wp,inidir);
  wp+=wcslen(inidir);
  wcscpy(wp,SUBDIR);
  wp+=wcslen(SUBDIR);

  CreateDirectoryW(
    ini,      // _In_     LPCTSTR               lpPathName,
    NULL      // _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes
  );

  *wp++=L'\\';
  *wp=0;

  // copy the basename without extension.
  memcpy(wp,pp,(ep-pp)*sizeof *wp);
  wp+=ep-pp;
  *wp=0;

  // copy the extension.
  wcscpy(wp,L"ini");

  /////////////////////////////////////////////////////////////////////////////
  code=0;
eout:
  return code;
}

int PlayerOpen(Player *pPlayer, int samplerate, int numchannels,
    int bitspersamp, int *pnMaxLatency, int *pnFramesDevice)
{
  /////////////////////////////////////////////////////////////////////////////
  if (EngineClientOpen(&pPlayer->engine,samplerate,numchannels,bitspersamp,
      pnMaxLatency,pnFramesDevice)<0) {
    DMESSAGE("opening engine");
    goto exit;
  }
  return 0;
exit:
  return -1;
}

void PlayerClose(Player *pPlayer)
{
}


int PlayerLock(Player *pPlayer )
{
  if (YASAPI_FSM_UP(pPlayer,ePlayerStateCriticalSection)<0)
    goto exit;

  EnterCriticalSection(&pPlayer->cs);
  ++pPlayer->nLockRecursion;

  return 0;
exit:
  return -1;
}

void PlayerUnlock(Player *pPlayer)
{
  if (0==pPlayer->nLockRecursion)
    DMESSAGE("unexpected lock recursion");

  --pPlayer->nLockRecursion;
  LeaveCriticalSection(&pPlayer->cs);
}

int PlayerLockDeep(Player *pPlayer, int nLockRecursion )
{
  while (0<nLockRecursion) {
    if (PlayerLock(pPlayer )<0)
      goto exit;

    --nLockRecursion;
  }

  return 0;
exit:
  return -1;
}

void PlayerUnlockDeep(Player *pPlayer, int *pnLockRecursion)
{
  int nLockRecursion=pPlayer->nLockRecursion;

  *pnLockRecursion=nLockRecursion;

  while (nLockRecursion--)
    PlayerUnlock(pPlayer);
}


int PlayerPause(Player *pPlayer, int nPause)
{
  return 0;
}

int PlayerIsPlaying(Player *pPlayer)
{
  return 0;
}

int PlayerAvailable(Player *pPlayer)
{
  return 0;
}

int PlayerOutputTime(Player *pPlayer)
{
  return 0;
}

int PlayerWrittenTime(Player *pPlayer)
{
  return 0;
}

int PlayerWrite(Player *pPlayer, BYTE *buf, int len)
{
  return 0;
}


void PlayerVolume(Player *pPlayer, int nVolume)
{
  if (0<=nVolume) {
    pPlayer->param.nVolume=nVolume;
    EngineClientVolume(&pPlayer->engine,pPlayer);
  }
}

const OptionsCommon *PlayerOptionsCommonConfig(Player *pPlayer)
{
  if (YASAPI_FSM_UP(pPlayer,ePlayerStateCache)<0)
    goto exit;

  return &pPlayer->cache.options.common.config;
exit:
  return &gcOptionsCommon;
}

const OptionsDevice *PlayerOptionsDeviceConfig(Player *pPlayer)
{
  if (YASAPI_FSM_UP(pPlayer,ePlayerStateCache)<0)
    goto exit;

  return &pPlayer->cache.selected.config.pCacheDevice->pMember->config;
exit:
  return &gcOptionsDevice;
}
