/*
 * yasapi_player_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int PlayerStateCriticalSectionEnter(Player *pPlayer)
{
  InitializeCriticalSection(&pPlayer->cs);
  pPlayer->nLockRecursion=0;
  ++pPlayer->fsm.pState;

  return 0;
}

static void PlayerStateCriticalSectionLeave(Player *pPlayer)
{
  --pPlayer->fsm.pState;
  DeleteCriticalSection(&pPlayer->cs);
}

///////////////////////////////////////////////////////////////////////////////
static int PlayerStateCacheEnter(Player *pPlayer)
{
  if (CacheCreate(&pPlayer->cache,pPlayer)<0) {
    DMESSAGE("Creating cache");
    goto exit;
  }

  ++pPlayer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void PlayerStateCacheLeave(Player *pPlayer)
{
  --pPlayer->fsm.pState;
  CacheDestroy(&pPlayer->cache);
}

///////////////////////////////////////////////////////////////////////////////
IPlayerState gcaIPlayerState[ePlayerStateCount]={
#define YASAPI_PLAYER_STATE(x) { \
  .pId=#x, \
  .Enter=PlayerState ## x ## Enter, \
  .Leave=PlayerState ## x ## Leave, \
},
#include <yasapi_player_state.h>
#undef YASAPI_PLAYER_STATE
};
