/*
 * yasapi_present_server_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int PresentServerStateEventEnter(PresentServer *pServer)
{
  // This is a MANUAL RESET event!
  if (!(pServer->hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void PresentServerStateEventLeave(PresentServer *pServer)
{
  --pServer->fsm.pState;
  CloseHandle(pServer->hEvent);
}

///////////////////////////////////////////////////////////////////////////////
static int PresentServerStateTimerEnter(PresentServer *pServer)
{
  if (!(pServer->hTimer=CreateWaitableTimerW(NULL,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void PresentServerStateTimerLeave(PresentServer *pServer)
{
  --pServer->fsm.pState;
  CloseHandle(pServer->hTimer);
}

///////////////////////////////////////////////////////////////////////////////
static int PresentServerStateTimerStartEnter(PresentServer *pServer)
{
  Player *pPlayer=pServer->pClient->pPlayer;
  const OptionsCommon *pOptions=PlayerOptionsCommonConfig(pPlayer);
  LONG lPeriod=pOptions->present.xPeriod+0.5;
  LARGE_INTEGER DueTime={ .QuadPart=10ll*1000ll*lPeriod };
  HANDLE hTimer=pServer->hTimer;
  BOOL bCode;

  if (pOptions->present.bEnabled) {
    bCode=SetWaitableTimer(
      hTimer,     // _In_           HANDLE           hTimer,
      &DueTime,   // _In_     const LARGE_INTEGER    *pDueTime,
      lPeriod,    // _In_           LONG             lPeriod,
      NULL,       // _In_opt_       PTIMERAPCROUTINE pfnCompletionRoutine,
      NULL,       // _In_opt_       LPVOID           lpArgToCompletionRoutine,
      FALSE       // _In_           BOOL             fResume
    );

    if (!bCode) {
      DVMESSAGE("starting waitable timer: 0x%04lx",GetLastError());
      goto exit;
    }
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void PresentServerStateTimerStartLeave(PresentServer *pServer)
{
  --pServer->fsm.pState;
  CancelWaitableTimer(pServer->hTimer);
}

///////////////////////////////////////////////////////////////////////////////
static int PresentServerStateLoopEndEnter(PresentServer *pServer)
{
  DMESSAGE("illegal attempt entering ePresentServerStateLoopEnd");

  return -1;
}

static void PresentServerStateLoopEndLeave(PresentServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int PresentServerStateErrorEnter(PresentServer *pServer)
{
  DMESSAGE("illegal attempt entering ePresentServerStateError");

  return -1;
}

static void PresentServerStateErrorLeave(PresentServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
IPresentServerState gcaIPresentServerState[ePresentServerStateCount]={
#define YASAPI_PRESENT_SERVER_STATE(x) { \
  .pId=#x, \
  .Enter=PresentServerState ## x ## Enter, \
  .Leave=PresentServerState ## x ## Leave, \
},
#include <yasapi_present_server_state.h>
#undef YASAPI_PRESENT_SERVER_STATE
};
